#-*- ruby -*-

#
# VFS for ruby - open anything with URI
# 
#   Copyright: Tatsuki Sugiura <sugi@nemui.org>
#   License: Ruby's original or GPL2
#   $LastChangedDate: 2003-10-13 10:05:51 +0900 (Mon, 13 Oct 2003) $
#   $LastChangedBy: sugi $
#

require 'uri'

Object.constants.member?("FileBuiltIn") or
  FileBuiltIn = File.clone

Object.constants.member?("DirBuiltIn") or 
  DirBuiltIn = Dir.clone

module VFS
  VERSION = 0.1
  HIERARCHY_SEPARATOR = "/"
#  SEPARATOR = HIERARCHY_SEPARATOR

  class File; end
  class Dir;  end
  module SchemeHandler; end

  module Error
    class InvalidURI  < StandardError; end
    class Unsupported < NotImplementedError; end
    class Unknown     < ScriptError; end
  end # Error

  module MethodGenerator
    private
    def gen_delegate_method(relay_to, *symbols)
      unless Array === relay_to
	relay_to = [relay_to]
      end
      relay_to[1] ||= "*arg, &block"
      relay_to[2] ||= relay_to[1]
      class_name, def_arg_str, relay_arg_str = relay_to
      symbols.each { |sym|
	module_eval %Q{
          def #{sym}(#{def_arg_str})
             #{class_name}.#{sym}(#{relay_arg_str})
	  end
	}
      }
    end

    def relay_path_to_file_builtin(*symbols)
      gen_delegate_method([:FileBuiltIn,
			  "*arg, &block",
			  "uri.to_s, *arg, &block"],
			  *symbols)
    end

    def relay_path_to_dir_builtin(*symbols)
      gen_delegate_method([:DirBuiltIn,
			  "*arg, &block",
			  "uri.to_s, *arg, &block"],
			  *symbols)
    end

    def relay_to_file_builtin(*symbols)
      gen_delegate_method(:FileBuiltIn, *symbols)
    end

    def relay_to_dir_builtin(*symbols)
      gen_delegate_method(:DirBuiltIn, *symbols)
    end

    def raise_unsupported(*symbols)
      symbols.each { |sym|
	module_eval %Q{
          def #{sym}(*arg, &block)
	    raise Error::Unsupported,
	      "method '#{sym}' for \#{self} is not supported yet." +
	      " please report bug or send patch!"
	  end
	}
      }
    end

    # convert; Class.moge -> Class.new.moge
    def instance_methods_to_class_methods(*symbols)
      symbols.each { |sym|
	module_eval %Q{
          class << self
	    def #{sym}(uri, *arg, &block)
	      #{self}.new(uri).#{sym}(*arg, &block)
	    end
	  end
	}
      }
    end

    def toplevel_methods_to_instance_methods(*symbols)
      symbols.each { |sym|
	module_eval %Q{
	  def #{sym}(*arg, &block)
	    VFS.#{sym}(self.uri.to_s, *arg, &block)
	  end
	}
      }
    end

    # Define class variable reader as class method like attr_reader.
    # ex;
    #  class Foo
    #     extend MethodGenerator
    #     @@bar = "this is foo's bar"
    #     class_var_reader :bar
    #  end
    #  puts Foo.bar  # => "this is foo's bar"
    def class_var_reader(*symbols)
      symbols.each { |sym|
	module_eval %Q{
          class << self
	    def #{sym}; @@#{sym}; end
	  end
	}
      }
    end
      
    # Define class variable writer as class method like attr_writer.
    # ex;
    #  class Foo
    #     extend MethodGenerator
    #     @@bar = "this is foo's bar"
    #     class_var_writer :bar
    #     def get_bar; @@bar; end
    #  end
    #  puts Foo.get_bar  # => "this is foo's bar"
    #  Foo.bar = "new"
    #  puts Foo.get_bar  # => "new"
    def class_var_writer(*symbols)
      symbols.each { |sym|
	module_eval <<-EOD
	  class << self
	    def #{sym}=(v); @@#{sym} = v; end
	  end
	EOD
      }
    end

    # Define class variable either reader and writer.
    def class_var_accessor(*symbols)
      class_var_reader(*symbols)
      class_var_writer(*symbols)
    end

    alias :class_attr_accessor :class_var_accessor
    alias :class_attr_reader   :class_var_reader
    alias :class_attr_writer   :class_var_writer

    def return_false(*symbols)
      symbols.each { |sym|
	module_eval "def #{sym}; false; end"
      }
    end

    def return_true(*symbols)
      symbols.each { |sym|
	module_eval "def #{sym}; true; end"
      }
    end

  end # MethodGenerator

  class Log
    extend MethodGenerator

    i = 1
    DEBUG    = (i+=1)
    INFO     = (i+=1)
    NOTICE   = (i+=1)
    WARNING  = (i+=1)
    ERROR    = (i+=1)
    CRIT     = (i+=1)
    @@loglevel = $DEBUG ? NOTICE : ERROR

    class_attr_accessor :loglevel
    
    class << self
      def log(msg, level = NOTICE)
	level >= loglevel and $stderr.puts msg
      end
    end
  end # Log

  # common methods has no relation to self
  module TopLevelMethods
    def loglevel
      Log.loglevel
    end

    def loglevel=(v)
      Log.loglevel=(v)
    end

    def join(*arg)
      arg.join(HIERARCHY_SEPARATOR)
    end

    # return schceme name
    def scheme(uri)
      if uri.kind_of?(URI)
	uri.scheme || "file"
      else
        URI.parse(uri).scheme || "file"
      end
    end

    def escape(s)
      # FIXME; replace by URI.escape
      s == nil and return nil
      str = s.dup
      # only generic escape!
      # this method may be overrided by each schemehandler???

      # FIXME? correct this. ref RFC!
      str.gsub(/[^ 0-9A-Za-z\-_.\/]+/) {|s|
        ('%%%02X' * s.size) % s.unpack('C*')
      }.tr(' ','+')
    end

    def unescape(s)
      # FIXME; replace by URI.unescape
      s == nil and return nil
      str = s.dup
      str.tr('+',' ').gsub(/(?:%[0-9a-zA-Z][0-9a-zA-Z])+/n) {|x|
        x.delete!('%')
        [x].pack('H*')
      }
    end
  end # TopLevelMethods
  extend VFS::TopLevelMethods

  # common methods has relation to self
  module Common
    def inspect
      %Q{<#{self.class} URI="#{uri}">}
    end

    private
    def log(msg, *level)
      msg = "VFS(#{self.class.to_s.split(/::/, 3)[2]}): #{msg}"
      Log.log(msg, *level)
    end
  end # Common
  
end # VFS
